/*
 * Broadcom GbE.
 * First cut, only the 5704, x86, little-endian.
 * Fix uint/int/u32int mess, return values.
 *
 * Csr32[rw] may need to become functions to deal with
 * chip idiosyncracies. Later.
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"
#include "../port/error.h"
#include "../port/netif.h"

#include "etherif.h"
#include "ethermii.h"

enum {					/* PCI configuration space */
	Mhc		= 0x68,		/* Miscellaneous Host Control */
	Dma		= 0x6C,		/* DMA Control */
	Ps		= 0x70,		/* PCI State */
	Pcc		= 0x74,		/* PCI Clock Control */
	Ria		= 0x78,		/* Register Indirect Address */
	Rid		= 0x80,		/* Register Indirect Data */
	Mwba		= 0x7C,		/* Memory Window Base Address */
	Mwd		= 0x84,		/* Memory Window Data */
	Dmc		= 0xB8,		/* Dual MAC Control */
};

enum {					/* Mhc */
	Is		= 0x00000001,	/* Interrupt State */
	Mpio		= 0x00000002,	/* Mask PCI Interrupt Output */
	Eebs		= 0x00000004,	/* Enable Endian Byte Swap */
	Eews		= 0x00000008,	/* Enable Endian Word Swap */
	Psrw		= 0x00000010,	/* PCI State R/W */
	Ia		= 0x00000080,	/* Indirect Access */
	CvMASK		= 0xFFFF0000,	/* Chip Version Mask */
	CvSHIFT		= 16,
	MetalMASK	= 0x00FF0000,	/* Metal */
	MetalSHIFT	= 16,
	ChipMASK	= 0x0F000000,	/* Chip */
	ChipSHIFT	= 24,
	AsicMASK	= 0xF0000000,	/* ASIC */
	AsicSHIFT	= 28,

	BCM5700		= 0x07,		/* ASIC */
	BCM5701		= 0x00,
	BCM5703		= 0x01,
	BCM5704		= 0x02,
	BCM5705		= 0x03,
	BCM5750		= 0x02,
};

enum {					/* Dma */
	DMAminMASK	= 0x000000FF,	/* Minimum DMA */
	DMAminSHIFT	= 0,
	DMA1dma		= 0x00004000,	/* One DMA at Once */
	DMArwMASK	= 0x00070000,	/* Read Watermark */
	DMArwSHIFT	= 16,
	DMAwwMASK	= 0x00380000,	/* Write Watermark */
	DMAwwSHIFT	= 19,
	DMArcmd		= 0x06000000,
	DMAwcmd		= 0x70000000,
};

enum {					/* Ps */
	Ispci		= 0x00000004,	/* Conventional PCI bus (!PCI-X) */
	Pci33		= 0x00000008,	/* PCI Bus Speed 33/66MHz (!66/133) */
	Bus32		= 0x00000010,	/* PCI Bus 32-bit (!64-bit) */
};

enum {					/* Registers */
	Imbox		= 0x0204,	/* Interrupt Mailbox (low 32-bits) */
	Rrpimbox	= 0x026C,	/* Receive Ring Producer Index */
	Rjrpimbox	= 0x0274,	/* Receive Jumbo Ring Producer Index */
	Rmrpimbox	= 0x027C,	/* Receive Mini Ring Producer Index */
	Rrrcimbox	= 0x0284,	/* Receive Return Ring Consumer Index */
	Shpimbox	= 0x0304,	/* Send Host Producer Index */
	Snpimbox	= 0x0384,	/* Send NIC Producer Index */
	MACmode		= 0x0400,	/* MAC mode */
	MACaddr		= 0x0410,	/* MAC Address 1 (8 bytes) */
	MACrbs		= 0x0438,	/* Random Backoff Seed */
	MACrmtu		= 0x043C,	/* Receive MTU size */
	MACmic		= 0x044C,	/* MI Communication */
	MACmim		= 0x0454,	/* MI Mode */
	MACtxl		= 0x0464,	/* Tx Lengths */
	MACrrc		= 0x0500,	/* Receive Rules Config. */
	Sdictl		= 0x0C08,	/* Send Data Initiator Control */
	Sdiena		= 0x0C0C,	/* Send Data Initiator Enable Mask */
	Rlpcfg		= 0x2010,	/* Receive List Placement Config. */
	Rlpctl		= 0x2014,	/* Receive List Placement Control */
	Rlpena		= 0x2018,	/* Receive List Placement Enable Mask */
	Rjrcb		= 0x2440,	/* Receive Jumbo Ring Rcb */
	Rrcb		= 0x2450,	/* Receive Ring Rcb */
	Rmrcb		= 0x2460,	/* Receive Mini Ring Rcb */
	Rrrt		= 0x2C18,	/* Receive Ring Repenish Threshold */
	HCCmode		= 0x3C00,	/* Host Coalescing Control Mode */
	HCCrt		= 0x3C08,	/* Receive Ticks */
	HCCtt		= 0x3C0C,	/* Transmit Ticks */
	HCCmrbd		= 0x3C10,	/* Max. Rbd's */
	HCCmsbd		= 0x3C14,	/* Max. Sbd's */
	HCCrit		= 0x3C18,	/* Receive Interrupt Ticks */
	HCCtit		= 0x3C1C,	/* Transmit Interrupt Ticks */
	HCCirbd		= 0x3C20,	/* Interrupt Rbd's */
	HCCisbd		= 0x3C24,	/* Interrupt Sbd's */
	HCCst		= 0x3C28,	/* Stat* Ticks */
	HCCstatsha	= 0x3C30,	/* Statistics Host Address (64-bit) */
	HCCstatusha	= 0x3C38,	/* Status Host Address (64-bit) */
	HCCstatsna	= 0x3C40,	/* Statistics NIC Address */
	HCCstatusna	= 0x3C44,	/* Status NIC Address */
	Marb		= 0x4000,	/* Memory Arbiter */
	BMmode		= 0x4400,	/* Buffer Manager (BM) Mode */
	BMba		= 0x4408,	/* BM Buffer pool Address */
	BMbsz		= 0x440C,	/* BM Buffer pool Size */
	BMblo		= 0x4410,	/* BM Buffer Low watermark */
	BMmlo		= 0x4414,	/* BM Buffer MAC Low watermark */
	BMbhi		= 0x4418,	/* BM Buffer High watermark */
	BMda		= 0x442C,	/* BM DMA Descriptor Address */
	BMdsz		= 0x4430,	/* BM DMA Descriptor Size */
	BMdlo		= 0x4434,	/* BM DMA Low watermark */
	BMdhi		= 0x4438,	/* BM DMA High watermark */
	FTQr		= 0x5C00,	/* Flow Through Queue Reset */
	Mctl		= 0x6800,	/* Mode Control */
	Mcfg		= 0x6804,	/* Miscellaneous Configuration */
	Mlc		= 0x6808,	/* Miscellaneous Local Control */
	NVRcmd		= 0x7000,	/* NVRAM Command */
	NVRaddr		= 0x700C,	/* NVRAM Address */
	NVRdata		= 0x7010,	/* NVRAM Data */
	Cfg1		= 0x7014,	/* NVRAM Configuration */
	Swarb		= 0x7020,	/* NVRAM Lock */
};

enum {					/* MACmode */
	PortHD		= 0x00000002,	/* Half Duplex */
	PortMASK	= 0x0000000C,	/* Active Port */
	PortMII		= 0x00000004,	/* 100Mb */
	PortGMII	= 0x00000008,	/* 1000Mb */
	PortTBI		= 0x0000000C,	/* Fiber */
	Rxstatse	= 0x00000800,	/*  */
	Rxstatsclr	= 0x00001000,	/*  */
	Txstatse	= 0x00004000,	/*  */
	Txstatsclr	= 0x00008000,	/*  */
	Txdmae		= 0x00200000,	/*  */
	Rxdmae		= 0x00400000,	/*  */
	Hdrdmae		= 0x00800000,	/*  */
};

enum {					/* MACmic */
	MidataMASK	= 0x0000FFFF,	/*  */
	MidataSHIFT	= 0,
	MiregMASK	= 0x001F0000,
	MiregSHIFT	= 16,
	MiphyMASK	= 0x03E00000,
	MiphySHIFT	= 21,
	Miwrite		= 0x04000000,
	Miread		= 0x08000000,
	Mifail		= 0x10000000,
	Mibusy		= 0x20000000,
	Mistart		= Mibusy,
};

enum {					/* MACmim */
	Ape		= 0x00000010,	/* Auto-Polling Enable */
};

enum {					/* Fmbox */
	Magic		= 0x4B657654,	/* */
};

enum {					/* Ncfg */
	Copper		= 0x00000010,	/* */
};

enum {					/* Marb */
	Marbrst		= 0x00000001,	/* Reset */
	Marbe		= 0x00000002,	/* Enable */
};

enum {					/* HCCmode */
	HCCrst		= 0x00000001,	/* Reset */
	HCCe		= 0x00000002,	/* Enable */
	HCCae		= 0x00000004,	/* Attention Enable */
	HCCnow		= 0x00000008,	/*  */
	HCC74byte	= 0x00000080,	/*  */
	HCC32byte	= 0x00000100,	/*  */
};

enum {					/* BMmode */
	BMrst		= 0x00000001,	/* Reset */
	BMe		= 0x00000002,	/* Enable */
	BMae		= 0x00000004,	/* Attention Enable */
	BMlae		= 0x00000010,	/* Low buffer Attention Enable */
};

enum {					/* Mctl */
	BswapBD		= 0x00000002,	/* Byte Swap Buffer Descriptors */
	WswapBD		= 0x00000004,	/* Word Swap Buffer Descriptors */
	BswapDMA	= 0x00000010,	/* Byte Swap DMA Data */
	WswapDMA	= 0x00000020,	/* Word Swap DMA Data */
	Stackup		= 0x00010000,	/*  */
	Hsb		= 0x00020000,	/* Host Send BDs */
	Txnopcksum	= 0x00100000,	/* No Tx Pseudoheader Checksum */
	Rxnopcksum	= 0x00800000,	/* No Rx Pseudoheader Checksum */
	Mai		= 0x04000000,	/* MAC Attention Interrupt */
};

enum {					/* Mcfg */
	Ccr		= 0x00000001,	/* Core Clock Reset */
	CcpMASK		= 0x000000FE,	/* Core Clock Prescalar */
	CcpSHIFT	= 1,
};

enum {					/* Mlc */
	Istate		= 0x00000001,	/* Interrupt State */
	Ionattn		= 0x00000008,
};

enum {					/* NVRcmd */
	NVRreset	= 0x00000001,
	NVRdone		= 0x00000008,
	NVRgo		= 0x00000010,
	NVRfirst	= 0x00000080,
	NVRlast		= 0x00000100,
};

enum {					/* Cfg1 */
	NVRfe		= 0x00000001,	/* Flash Enabled */
	NVRbm		= 0x00000002,	/* Buffered Mode */
	NVRcb		= 0x80000000,	/* Compatibility Bypass */

	NVRbmpsz	= 264,		/* Buffered Mode Page Size */
	NVRbmpshft	= 9,		/* BUffered Mode Page Shift */
};

enum {					/* Swarb */
	Set1		= 0x00000002,
	Clr1		= 0x00000020,
	Gnt1		= 0x00000200,
};

enum {					/* NIC SRAM Memory Window */
	Msrcb		= 0x0100,	/* Send Rcb (*16) */
	Mrrrcb		= 0x0200,	/* Receive Return Rcb (*16) */
	Mstats		= 0x0300,	/* Stats Block */
	Mstatus		= 0x0B00,	/* Status Block */
	Mfmbox		= 0x0B50,	/* Firmware Mailbox */
	Mncfgsig	= 0x0B54,	/* NIC Configuration Signature */
	Mncfg		= 0x0B58,	/* NIC Configuration */
	Mmmbox		= 0x0C14,	/* MAC mailbox (MAC high address) */
	Mbmda		= 0x2000,	/* Buffer Manager DMA Descriptors */
	Msr		= 0x4000,	/* Send Ring */
	Mrr		= 0x6000,	/* Receive Ring */
	Mbmba		= 0x8000,	/* Buffer Manager Buffers */
};

typedef struct Host64 {
	u32int	hi;
	u32int	lo;
} Host64;

typedef struct Rbd {			/* Receive Buffer Descriptor */
	Host64	addr;
	u32int	indexlen;		/* (ring-index<<16)|buffer-length */
	u32int	flags;			/* only lower 16-bits */
	u32int	checksum;		/* (ip<<16)|tcp/udp */
	u32int	error;			/* only upper 16-bits */
	u32int	reserved;
	void*	opaque;			/* passed to receive return ring */
} Rbd;

typedef struct Sbd {			/* Send Buffer Descriptor */
	Host64	addr;
	u32int	lenflags;		/* (len<<16)|flags */
	u32int	reserved;
} Sbd;

/*
 * Unlike the Tigon II (e.g. in the GA620), the ring control
 * blocks are held in registers and not host memory, so there's
 * no Gib and no need to keep the control blocks in the Ctlr.
 */
typedef struct Rcb {			/* Ring Control Block */
	Host64	ring;			/* points to the Rbd ring */
	u32int	control;		/* (max_len<<16)|flags */
	u32int	addr;			/* address in SRAM */
} Rcb;

enum {					/* Rcb */
	MaxlenMASK	= 0xFFFF0000,
	MaxlenSHIFT	= 16,
	RCBdisabled	= 0x0002,	/* Jumbo or Mini RCB only */
};

#ifdef notdef
typedef struct Gib {			/* General Information Block */
	uchar	statistics[2048];	/* Statistics (TBD) */
	Rcb	rrcb;			/* Receive Ring */
} Gib;
#endif /* notdef */

typedef struct Stats {
	uchar	statistics[2048];	/* Statistics (TBD) */
} Stats;

typedef struct Status {
	u32int	status;
	u32int	unused[3];
	struct {
		u16int	rpi;
		u16int	sci;
	} i[16];
} Status;

enum {					/* Host/NIC Interface ring sizes */
	Ner		= 256,		/* Event Ring */
	Ncr		= 64,		/* Command Ring */
	Nsr		= 512,		/* Send Ring */
	Nrr		= 512,		/* Receive Ring */
	Nrrr		= 1024,		/* Receive Return Ring */
};

typedef struct Ctlr Ctlr;
typedef struct Ctlr {
	int	port;
	Pcidev*	pcidev;
	Ctlr*	next;
	int	active;
	uint	id;
	uchar	ea[Eaddrlen];

	u32int*	nic;

	QLock	alock;			/* attach */
	void*	alloc;			/*  */
	int	cls;

	Lock	clock;			/* controller registers */

	uint	mhc;			/* Mhc */
	uint	asic;
	uint	chip;

	uint	flags;

	Rbd*	rr;			/* Receive Ring */
	int	nrr;			/* currently in Receive Ring */
	Sbd*	sr;			/* Send Ring */
	int	nsr;			/* currently in Send Ring */
	Rbd*	rrr;			/* Receive Return Ring */
	int	rrrci;			/* Receive Return Ring Consumer Index */

	int	spi;			/* send producer index */
	int	sci;			/* send consumer index */

	Mii*	mii;

	Stats*	stats;			/* statistics */
	Status*	status;

	uint	intr;
	uint	lintr;			
	uint	lsleep;
	uint	rintr;
	uint	tintr;
} Ctlr;

enum {
	Feeprom		= 0x00000001,	/* use EEPROM for configuration info */
	Fnvr		= 0x00000002,	/* use NVRAM for configuration info */
	Fnvrinit	= 0x00000004,	/* have done NVRAM initialisation */
	Fnvrbm		= 0x00000008,	/* NVRAM Buffered Mode */
	Fpcie		= 0x00000010,	/* PCI Express */
};

static Ctlr* bgbectlrhead;
static Ctlr* bgbectlrtail;

#define csr32r(c, r)	(*((c)->nic+((r)/4)))
#define csr32w(c, r, v)	(*((c)->nic+((r)/4)) = (v))

static void
sethost64(Host64* host64, void* addr)
{
	uvlong uvl;

	uvl = PCIWADDR(addr);
	host64->hi = uvl>>32;
	host64->lo = uvl & 0xFFFFFFFFL;
}

static long
bgbeifstat(Ether* edev, void* a, long n, ulong offset)
{
	char *p;
	Ctlr *ctlr;
	int i, l, r;

	ctlr = edev->ctlr;

	p = malloc(2*READSTR);
	l = 0;
	l += snprint(p+l, 2*READSTR-l, "intr: %ud\n", ctlr->intr);
	l += snprint(p+l, 2*READSTR-l, "lintr: %ud\n", ctlr->lintr);
	l += snprint(p+l, 2*READSTR-l, "lsleep: %ud\n", ctlr->lsleep);
	l += snprint(p+l, 2*READSTR-l, "rintr: %ud\n", ctlr->rintr);
	l += snprint(p+l, 2*READSTR-l, "tintr: %ud\n", ctlr->tintr);

	if(ctlr->mii != nil && ctlr->mii->curphy != nil){
		l += snprint(p+l, 2*READSTR, "phy:   ");
		for(i = 0; i < NMiiPhyr; i++){
			if(i && ((i & 0x07) == 0))
				l += snprint(p+l, 2*READSTR-l, "\n       ");
			r = miimir(ctlr->mii, i);
			l += snprint(p+l, 2*READSTR-l, " %4.4ux", r);
		}
		snprint(p+l, 2*READSTR-l, "\n");
	}
	snprint(p+l, 2*READSTR-l, "\n");

	n = readstr(offset, a, n, p);
	free(p);

	return n;
}

static void
bgbepromiscuous(void* arg, int on)
{
	USED(arg, on);
}

static void
bgbemulticast(void* arg, uchar* addr, int on)
{
	USED(arg, addr, on);
}

static void
bgbeattach(Ether* edev)
{
	Ctlr *ctlr;

	ctlr = edev->ctlr;
	qlock(&ctlr->alock);
	if(ctlr->alloc != nil){
		qunlock(&ctlr->alock);
		return;
	}

	qunlock(&ctlr->alock);
	poperror();
}

static void
bgbetransmit(Ether* edev)
{
	Ctlr *ctlr;

	ctlr = edev->ctlr;
	USED(ctlr);

	panic("bgbetransmit\n");
}

static void
bgbeinterrupt(Ureg*, void* arg)
{
	Ctlr *ctlr;
	Ether *edev;

	edev = arg;
	ctlr = edev->ctlr;

	ctlr->intr++;

	panic("bgbeinterrupt\n");
}

static int
bgbemiimir(Mii* mii, int pa, int ra)
{
	int timeo;
	u32int ape, r;
	Ctlr *ctlr;

	/*
	 * PHY is always at address 1, some
	 * chips respond to all addresses.
	 */
	if(pa != 1)
		return -1;
	ctlr = mii->ctlr;

	/*
	 * Turn off Auto Polling, it can cause problems during access.
	 */
	ape = csr32r(ctlr, MACmim);
	if(ape & Ape){
		csr32w(ctlr, MACmim, ape & ~Ape);
		microdelay(40);
	}

	r = Miread|(pa<<MiphySHIFT)|(ra<<MiregSHIFT);
	csr32w(ctlr, MACmic, Mistart|r);
	for(r = timeo = 0; timeo < 10000; timeo++){
		if(!((r = csr32r(ctlr, MACmic)) & Mibusy))
			break;
	}
	if(ape & Ape)
		csr32w(ctlr, MACmim, ape);
//print("miir: pa %d ra %d r %#8.8ux\n", pa, ra, r);

	if(r & Mibusy)
		return -1;
	return r & MidataMASK;
}

static int
bgbemiimiw(Mii* mii, int pa, int ra, int data)
{
	int timeo;
	u32int ape, r;
	Ctlr *ctlr;

	/*
	 * PHY is always at address 1, some
	 * chips respond to all addresses.
	 */
	if(pa != 1)
		return -1;
	ctlr = mii->ctlr;

	/*
	 * Turn off Auto Polling, it can cause problems during access.
	 */
	ape = csr32r(ctlr, MACmim);
	if(ape & Ape){
		csr32w(ctlr, MACmim, ape & ~Ape);
		microdelay(40);
	}

	r = Miwrite|(pa<<MiphySHIFT)|(ra<<MiregSHIFT)|(data & MidataMASK);
	csr32w(ctlr, MACmic, Mistart|r);
	for(r = timeo = 0; timeo < 10000; timeo++){
		if(!((r = csr32r(ctlr, MACmic)) & Mibusy))
			break;
	}
	if(ape & Ape)
		csr32w(ctlr, MACmim, ape);
//print("miiw: pa %d ra %d r %#8.8ux\n", pa, ra, r);

	if(r & Mibusy)
		return -1;
	return 0;
}

static int
mem32r(Ctlr* ctlr, int o)
{
	u32int r;

	ilock(&ctlr->clock);
	pcicfgw32(ctlr->pcidev, Mwba, o);
	r = pcicfgr32(ctlr->pcidev, Mwd);
	pcicfgw32(ctlr->pcidev, Mwba, 0);
	iunlock(&ctlr->clock);

	return r;
}

static void
mem32w(Ctlr* ctlr, int o, int data)
{
	ilock(&ctlr->clock);
	pcicfgw32(ctlr->pcidev, Mwba, o);
	pcicfgw32(ctlr->pcidev, Mwd, data);
	pcicfgw32(ctlr->pcidev, Mwba, 0);
	iunlock(&ctlr->clock);
}

static int
nvrcfgr(Ctlr* ctlr, int o)
{
	int cfg1, data, r, timeo;

	if(!(ctlr->flags & Fnvr) || (o & 3))
		return -1;

	/*
	 * Lock.
	 */
	csr32w(ctlr, Swarb, Set1);
	for(r = timeo = 0; timeo < 10000; timeo++){
		if((r = csr32r(ctlr, Swarb)) & Gnt1)
			break;
		microdelay(10);
	}
	if(!(r & Gnt1))
		return -1;

	if(!(ctlr->flags & Fnvrinit)){
		cfg1 = csr32r(ctlr, Cfg1);
		if(!(cfg1 & NVRfe))
			csr32w(ctlr, Cfg1, cfg1 & ~NVRcb);
		else
			ctlr->flags |= Fnvrbm;
		ctlr->flags |= Fnvrinit;
	}

	if(ctlr->flags & Fnvrbm)
		o = ((o/NVRbmpsz)<<NVRbmpshft) + (o % NVRbmpsz);

	csr32w(ctlr, NVRaddr, o);
	csr32w(ctlr, NVRcmd, NVRlast|NVRfirst|NVRgo);

	data = -1;
	for(r = timeo = 0; timeo < 1000; timeo++){
		microdelay(10);
		if((r = csr32r(ctlr, NVRcmd)) & NVRdone){
			microdelay(10);
			data = csr32r(ctlr, NVRdata);
			break;
		}
	}

	/*
	 * Unlock.
	 */
	csr32w(ctlr, Swarb, Clr1);

	if(!(r & NVRdone))
		return -1;

	return data;
}

static int
bgbeinit(Ctlr* ctlr)
{
	uint r;
	Host64 host64;
	int i, timeo;
	uchar *alloc;

	/*
	 * This will be in attach,
	 * but here now for development because I can't be
	 * bothered waiting to log in to find out if this stuff
	 * works.
	 * Need to alloc enough for
	 *	 sizeof(Rbd)*Nrr
	 *	+sizeof(Sbd)*Nsr
	 *	+sizeof(Rbd)*Nrrr
	 *	+sizeof(Stats)
	 *	+sizeof(Status)
	 */
	if(ctlr->alloc != nil)
		return -1;
	alloc = malloc(sizeof(Rbd)*(Nrr+Nrrr)
			+sizeof(Rbd)*Nsr
			+sizeof(Stats)
			+sizeof(Status)
			+ctlr->cls-1);
	if(alloc == nil)
		return -1;
	ctlr->alloc = alloc;

	/*
	 * Stackup before this?
	 */
	/*
	 * Buffer Manager buffer and descriptor pools
	 * - address, size, low and high watermarks.
	 * Not the BCM5705/BCM5750.
	 */
	csr32w(ctlr, BMba, Mbmba);
	if(ctlr->asic == BCM5704)
		csr32w(ctlr, BMbsz, 0x10000);
	else
		csr32w(ctlr, BMbsz, 0x18000);
	csr32w(ctlr, BMblo, 80);
	csr32w(ctlr, BMmlo, 32);
	csr32w(ctlr, BMbhi, 96);

	csr32w(ctlr, BMda, Mbmda);
	csr32w(ctlr, BMdsz, 0x2000);
	csr32w(ctlr, BMdlo, 5);
	csr32w(ctlr, BMdhi, 10);

	/*
	 * Enable the Buffer Manager and wait for it
	 * to say it is. (Weird, the BNe bit is set and the
	 * loop tests the bit is set...).
	 * Note, the Broadcom and Linux drivers disagree with
	 * the FreeBSD driver as to whether the BMae or BMlae
	 * bit is also set.
	 */
	csr32w(ctlr, BMmode, BMae|BMe);
	for(r = timeo = 0; timeo < 10000; timeo++){
		if((r = csr32r(ctlr, BMmode)) & BMe)
			break;
		microdelay(10);
	}
	if(!(r & BMe))
		return -1;

	/*
	 * Enable flow-through queues.
	 */
	csr32w(ctlr, FTQr, 0xFFFFFFFF);
	csr32w(ctlr, FTQr, 0);
	for(timeo = 0; timeo < 10000; timeo++){
		if(!csr32r(ctlr, FTQr))
			break;
		microdelay(10);
	}
	if(csr32r(ctlr, FTQr))
		return -1;

	alloc = (uchar*)ROUNDUP((ulong)ctlr->alloc, ctlr->cls);

	/*
	 * Receive (Standard) Ring.
	 * Unlike the Tigon II (e.g. in the GA620), the ring control
	 * block is held in registers and not host memory, so there's
	 * no need to keep the control block in the Ctlr.
	 * The RSR replenish threshold is recommended to be 1/8 of the
	 * total ring.
	 *
	 * Jumbo and mini rings disabled.
	 */
	ctlr->rr = (Rbd*)alloc;
	alloc += sizeof(Rbd)*Nrr;
	sethost64(&host64, ctlr->rr);
	csr32w(ctlr, Rrcb+offsetof(Rcb, ring.hi), host64.hi);
	csr32w(ctlr, Rrcb+offsetof(Rcb, ring.lo), host64.lo);
	csr32w(ctlr, Rrcb+offsetof(Rcb, control), (ETHERMAXTU+4)<<MaxlenSHIFT);
	csr32w(ctlr, Rrcb+offsetof(Rcb, addr), Mrr);

	csr32w(ctlr, Rrrt, Nrr/8);

	csr32w(ctlr, Rjrcb+offsetof(Rcb, control), RCBdisabled);
	csr32w(ctlr, Rmrcb+offsetof(Rcb, control), RCBdisabled);

	csr32w(ctlr, Rrpimbox, 0);
	csr32w(ctlr, Rjrpimbox, 0);
	csr32w(ctlr, Rmrpimbox, 0);

	/* indices? */

	/*
	 * Send Ring.
	 * Host is the producer and updates Shpimbox
	 * when something to be sent (.spi is our local copy).
	 * NIC is consumer and updates the status block, the host has to
	 * then update .sci and free completed packets.
	 * Some chipsets cause problems by re-ordering writes to the mailbox
	 * registers, so read back. Why isn't this a problem for mailboxes
	 * other than Snpimbox?
	 * The send descriptors can be in either host or NIC memory;
	 * host is chosen here.
	 * Only one is used, the others disabled.
	 */
	ctlr->spi = ctlr->sci = 0;
	csr32w(ctlr, Shpimbox, 0);
	csr32w(ctlr, Snpimbox, 0);

	ctlr->sr = (Sbd*)alloc;
	alloc += sizeof(Sbd)*Nsr;
	sethost64(&host64, ctlr->sr);
	mem32w(ctlr, Msrcb+offsetof(Rcb, ring.hi), host64.hi);
	mem32w(ctlr, Msrcb+offsetof(Rcb, ring.lo), host64.lo);
	mem32w(ctlr, Msrcb+offsetof(Rcb, control), Nsr<<16);
	mem32w(ctlr, Msrcb+offsetof(Rcb, addr), Msr);
	for(i = Msrcb+sizeof(Rcb); i < Msrcb+sizeof(Rcb)*16; i += sizeof(Rcb))
		mem32w(ctlr, i+offsetof(Rcb, control), RCBdisabled);

	/*
	 * Receive Return Ring.
	 */
	ctlr->rrr = (Rbd*)alloc;
	alloc += sizeof(Rbd)*Nrrr;
	sethost64(&host64, ctlr->rrr);
	mem32w(ctlr, Mrrrcb+offsetof(Rcb, ring.hi), host64.hi);
	mem32w(ctlr, Mrrrcb+offsetof(Rcb, ring.lo), host64.lo);
	mem32w(ctlr, Mrrrcb+offsetof(Rcb, control), Nrrr<<16);
	mem32w(ctlr, Mrrrcb+offsetof(Rcb, addr), 0);
	for(i = Mrrrcb+sizeof(Rcb); i < Mrrrcb+sizeof(Rcb)*16; i += sizeof(Rcb))
		mem32w(ctlr, i+offsetof(Rcb, control), RCBdisabled);

	ctlr->rrrci = 0;
	csr32w(ctlr, Rrrcimbox, 0);

	/*
	 * Load MAC address.
	 * The chip has more than one, should they be
	 * initialised, how are they enabled/used?
	 */
	r = (ctlr->ea[0]<<8)|ctlr->ea[1];
	csr32w(ctlr, MACaddr, r);
	r = (ctlr->ea[2]<<24)|(ctlr->ea[3]<<16)|(ctlr->ea[4]<<8)|ctlr->ea[4];
	csr32w(ctlr, MACaddr+4, r);

	/*
	 * Initialise transmit random backoff.
	 * Note, the Broadcom and Linux drivers disagree with
	 * the FreeBSD driver as to the value of the mask.
	 */
	r = 0;
	for(i = 0; i < Eaddrlen; i++)
		r += ctlr->ea[0];
	csr32w(ctlr, MACrbs, r & 0x3FF);

	/*
	 * Various parameters:
	 *    MACrmtu: receive MTU - MTU + ethernet header + FCS.
	 *    MACtxl: inter-packet gap - change later if Gb+HD.
	 *    MACrrc: ring to use for packets that don't match any RX rules.
	 *    Rlpcfg: One interrupt distribution list, sixteen active lists,
	 *	one bad frames class.
	 *    Rlpena: stats mask
	 *    Rlpctl: and enable
	 *    Sdiena: stats mask
	 *    Sdictl: and enable and update
	 */
	csr32w(ctlr, MACrmtu, ETHERMAXTU+4);
	csr32w(ctlr, MACtxl, 0x2620);
	csr32w(ctlr, MACrrc, 0x08);
	csr32w(ctlr, Rlpcfg, 0x181);
	csr32w(ctlr, Rlpena, 0x007FFFFF);
	csr32w(ctlr, Rlpctl, 0x01);
	csr32w(ctlr, Sdiena, 0x00FFFFFF);
	csr32w(ctlr, Sdictl, 0x03);

	/*
	 * Host Coalescing Control.
	 * Disable it then set up the defaults.
	 * Values vary between drivers:
	 *			bsd	linux	broadcom
	 *	HCCrt		150	   0	   60
	 *	HCCtt		150	  96	  200 
	 *	HCCmrbd		 64	   1	   15
	 *	HCCmsbd		128	   5	   35
	 *	HCCrit		  0	   0	   25
	 *	HCCtit		  0	   0	   25
	 *	HCCirbd		  0	   1	    2
	 *	HCCisbd		  0	   0	    5
	 * The Broadcom driver has adaptive high/low limits.
	 * Put in tuning control later.
	 */
	csr32w(ctlr, HCCmode, 0);
	for(r = timeo = 0; timeo < 10000; timeo++){
		if(!((r = csr32r(ctlr, HCCmode)) & HCCe))
			break;
		microdelay(10);
	}
	if(r & HCCe)
		return -1;

	csr32w(ctlr, HCCrt, 150);
	csr32w(ctlr, HCCtt, 150);
	csr32w(ctlr, HCCmrbd, 64);
	csr32w(ctlr, HCCmsbd, 128);
	if(ctlr->asic != BCM5705){
		csr32w(ctlr, HCCrit, 0);
		csr32w(ctlr, HCCtit, 0);
	}
	csr32w(ctlr, HCCirbd, 0);
	csr32w(ctlr, HCCisbd, 0);

	/*
	 * Statistics/Status.
	 */
	ctlr->stats = (Stats*)alloc;
	alloc += sizeof(Stats);
	sethost64(&host64, ctlr->stats);
	csr32w(ctlr, HCCstatsha, host64.hi);
	csr32w(ctlr, HCCstatsha+4, host64.lo);
	csr32w(ctlr, HCCstatsha, Mstats);

	ctlr->status = (Status*)alloc;
	alloc += sizeof(Status);
	sethost64(&host64, ctlr->status);
	csr32w(ctlr, HCCstatusha, host64.hi);
	csr32w(ctlr, HCCstatusha+4, host64.lo);
	csr32w(ctlr, HCCstatusha, Mstatus);
	ctlr->status->i[0].rpi = ctlr->status->i[0].sci = 0;

	csr32w(ctlr, HCCst, 1000000);

	/*
	 * All set up, now start to enable stuff:
	 *	host coalescing state machine;
	 */
	csr32w(ctlr, HCCmode, HCC32byte|HCCe);

	/* to be continued... */
#ifdef notdef
/* Receive BD Completion Control Registers */
#define RCVCC_MODE			0x00003000
#define  RCVCC_MODE_RESET		 0x00000001
#define  RCVCC_MODE_ENABLE		 0x00000002
#define  RCVCC_MODE_ATTN_ENABLE		 0x00000004
	tw32(RCVCC_MODE, RCVCC_MODE_ENABLE | RCVCC_MODE_ATTN_ENABLE);
or
#define BGE_RBDC_MODE			0x3000
	/* Turn on RX BD completion state machine and enable attentions */
	CSR_WRITE_4(sc, BGE_RBDC_MODE,
	    BGE_RBDCMODE_ENABLE|BGE_RBDCMODE_ATTN);



	/* Turn on RX list placement state machine */
	CSR_WRITE_4(sc, BGE_RXLP_MODE, BGE_RXLPMODE_ENABLE);

	/* Turn on RX list selector state machine. */
	if (sc->bge_asicrev != BGE_ASICREV_BCM5705)
		CSR_WRITE_4(sc, BGE_RXLS_MODE, BGE_RXLSMODE_ENABLE);

	/* Turn on DMA, clear stats */
	CSR_WRITE_4(sc, BGE_MAC_MODE, BGE_MACMODE_TXDMA_ENB|
	    BGE_MACMODE_RXDMA_ENB|BGE_MACMODE_RX_STATS_CLEAR|
	    BGE_MACMODE_TX_STATS_CLEAR|BGE_MACMODE_RX_STATS_ENB|
	    BGE_MACMODE_TX_STATS_ENB|BGE_MACMODE_FRMHDR_DMA_ENB|
	    (sc->bge_tbi ? BGE_PORTMODE_TBI : BGE_PORTMODE_MII));

	/* Set misc. local control, enable interrupts on attentions */
	CSR_WRITE_4(sc, BGE_MISC_LOCAL_CTL, BGE_MLC_INTR_ONATTN);
#endif /* notdef */
USED(alloc);
print("bgbeinit ok\n");

	return 0;
}

static int
bgbedetach(Ctlr* ctlr)
{
	int pcr, timeo;

	/*
	 * Soft reset the controller.
	 * Pcr and Cls must be saved and restored after the reset,
	 * as must Mhc.
	 * The magic number written to the mailbox will be complemented
	 * by the firmware once initialisation is complete.
	 */
	pcr = pcicfgr16(ctlr->pcidev, PciPCR);
	mem32w(ctlr, Mfmbox, Magic);

	csr32w(ctlr, Mcfg, (65<<CcpSHIFT)|Ccr);
	delay(1000);

	pcicfgw8(ctlr->pcidev, PciCLS, ctlr->cls);
	pcicfgw16(ctlr->pcidev, PciPCR, pcr);
	pcicfgw32(ctlr->pcidev, Mhc, Ia|Psrw|Eews|Mpio|Is);

	/*
	 * Wait for the firmware to become ready.
	 */
	for(timeo = 0; timeo < 100000; timeo++){
		if(mem32r(ctlr, Mfmbox) == ~Magic)
			break;
		microdelay(10);
	}
	if(timeo >= 1000)
		return -1;

	csr32w(ctlr, Marb, Marbe);
	csr32w(ctlr, MACmode, 0);

	return 0;
}

static int
bgbereset(Ctlr* ctlr)
{
	int x;
	MiiPhy *phy;
	u32int dma, ps, r;

	if(bgbedetach(ctlr) < 0)
		return -1;

	/*
	 * Set up PCI DMA parameters.
	 * Different for regular PCI and PCI-X.
	 * PCI Express is another matter.
	 */
	dma = DMAwcmd|DMArcmd;
	ps = pcicfgr32(ctlr->pcidev, Ps);
	if(ps & Pci33){
		dma |= (0x7<<DMAwwSHIFT)|(0x7<<DMArwSHIFT);
		if(ctlr->asic != BCM5705 && ctlr->asic != BCM5750)
			dma |= 0x0F;
	}
	else{
		if(ctlr->asic == BCM5704)
			dma |= (0x3<<DMAwwSHIFT)|(0x7<<DMArwSHIFT);
		else
			dma |= (0x3<<DMAwwSHIFT)|(0x3<<DMArwSHIFT)|0x0F;

		/*
		 * Hardware bug workarounds.
		 */
		if(ctlr->asic == BCM5703 || ctlr->asic == BCM5704){
			r = csr32r(ctlr, Pcc) & 0x1F;
			if(r == 0x06 || r == 0x07)
				dma |= DMA1dma;
			dma |= 0x00800000;
		}
	}
	if(ctlr->asic == BCM5703 || ctlr->asic == BCM5704 || ctlr->asic == BCM5705)
		dma &= ~DMAminMASK;
	pcicfgw32(ctlr->pcidev, Dma, dma);

	/*
	 * Doesn't work, apparently.
	 */
	pciclrmwi(ctlr->pcidev);

	/*
	 * Set up general mode register
	 * and timer pre-scalar.
	 */
	csr32w(ctlr, Mctl, Mai|Rxnopcksum|Txnopcksum|Hsb|WswapBD|BswapDMA|WswapDMA);
	csr32w(ctlr, Mcfg, (65<<CcpSHIFT));

	r = mem32r(ctlr, Mmmbox);
	if((r & 0xFFFF0000) == 0x484B0000){
		ctlr->ea[0] = (r>>8) & 0xFF;
		ctlr->ea[1] = r & 0xFF;
		r = mem32r(ctlr, Mmmbox+4);
		ctlr->ea[2] = (r>>24) & 0xFF;
		ctlr->ea[3] = (r>>16) & 0xFF;
		ctlr->ea[4] = (r>>8) & 0xFF;
		ctlr->ea[5] = r & 0xFF;
	}
	else if(ctlr->flags & Fnvr){
		if(pcicfgr32(ctlr->pcidev, Dmc) & 0x04)
			x = 0xCC;
		else
			x = 0x7C;
		r = nvrcfgr(ctlr, x);
		ctlr->ea[0] = (r>>8) & 0xFF;
		ctlr->ea[1] = r & 0xFF;
		r = nvrcfgr(ctlr, x+4);
		ctlr->ea[2] = (r>>24) & 0xFF;
		ctlr->ea[3] = (r>>16) & 0xFF;
		ctlr->ea[4] = (r>>8) & 0xFF;
		ctlr->ea[5] = r & 0xFF;
	}
	else{
		/*
		 * Read the EEPROM or something. Later.
		 */
		/*nothing to do*/;
	}

	/*
	 * Not interested in getting the hardware configuration
	 * from the EEPROM or in anything other than twisted-pair
	 * right now.
	 */
	if(mem32r(ctlr, Mncfgsig) != Magic)
		return -1;
	if(!(mem32r(ctlr, Mncfg) & Copper))
		return -1;

bgbeinit(ctlr);

	/*
	 * Link management.
	 */
	if((ctlr->mii = malloc(sizeof(Mii))) == nil)
		return -1;
	ctlr->mii->mir = bgbemiimir;
	ctlr->mii->miw = bgbemiimiw;
	ctlr->mii->ctlr = ctlr;

	if(mii(ctlr->mii, ~0) == 0 || (phy = ctlr->mii->curphy) == nil){
		free(ctlr->mii);
		ctlr->mii = nil;
		return -1;
	}
	print("oui %#ux phyno %d\n", phy->oui, phy->phyno);

	miiane(ctlr->mii, ~0, ~0, ~0);
delay(3500);
miistatus(ctlr->mii);
r = miimir(ctlr->mii, Bmsr);
print("bmsr %#4.4ux anar %#4.4ux fc %#4.4ux mscr %#4.4ux\n",
  r, ctlr->mii->curphy->anar,
  ctlr->mii->curphy->fc, ctlr->mii->curphy->mscr);
print("link %d speed %d fd %d rfc %d tfc %d\n",
  ctlr->mii->curphy->link, ctlr->mii->curphy->speed,
  ctlr->mii->curphy->fd, ctlr->mii->curphy->rfc, ctlr->mii->curphy->rfc);

	return 0;
}

static void
bgbepci(void)
{
	int port;
	ulong bar;
	Pcidev *p;
	Ctlr *ctlr;

	p = nil;
	while(p = pcimatch(p, 0, 0)){
		if(p->ccrb != 0x02 || p->ccru != 0)
			continue;

		switch((p->did<<16)|p->vid){
		default:
		case (0x1644<<16)|0x14E4:	/* BCM5700 */
		case (0x1645<<16)|0x14E4:	/* BCM5701 */
		case (0x1646<<16)|0x14E4:	/* BCM5702 */
		case (0x1647<<16)|0x14E4:	/* BCM5703 */
			continue;
		case (0x1648<<16)|0x14E4:	/* BCM5704 */
			break;
		}

		bar = p->mem[0].bar;
		if(bar & 0x01){
			port = bar & ~0x01;
			if(ioalloc(port, p->mem[0].size, 0, "bgbe") < 0){
				print("bgbe: port %#ux in use\n", port);
				continue;
			}
		}
		else{
			port = upamalloc(bar & ~0x0F, p->mem[0].size, 0);
			if(port == 0){
				print("bgbe: can't map %#8.8lux\n", bar);
				continue;
			}
		}
		ctlr = malloc(sizeof(Ctlr));
		ctlr->port = port;
		ctlr->pcidev = p;
		ctlr->id = (p->did<<16)|p->vid;

		ctlr->nic = KADDR(ctlr->port);
		ctlr->cls = pcicfgr8(p, PciCLS);

		/*
		 * Buggered.
		 * PCI routines should have an extra argument which is
		 * where to place the unsigned data, and the the return
		 * value signifies success or failure.
		 */
		ctlr->mhc = (pcicfgr32(p, Mhc) & CvMASK)|Ia|Psrw|Eews|Mpio|Is;
		ctlr->asic = (ctlr->mhc & AsicMASK)>>AsicSHIFT;
		ctlr->chip = (ctlr->mhc & ChipMASK)>>ChipSHIFT;
		switch(ctlr->asic){
		default:
			free(ctlr);
			continue;
		case BCM5703:
		case BCM5704:
			ctlr->flags |= Fnvr;
			break;
		}

		/*
		 * Need to set this before poking at the MMIO regs.
		 */
		pcicfgw32(ctlr->pcidev, Mhc, ctlr->mhc);
		pcicfgw32(ctlr->pcidev, Mwba, 0);
		if(bgbereset(ctlr)){
			free(ctlr);
			continue;
		}
		pcisetbme(p);

		if(bgbectlrhead != nil)
			bgbectlrtail->next = ctlr;
		else
			bgbectlrhead = ctlr;
		bgbectlrtail = ctlr;
	}
}

static int
bgbepnp(Ether* edev)
{
	Ctlr *ctlr;

	if(bgbectlrhead == nil)
		bgbepci();

	/*
	 * Any adapter matches if no edev->port is supplied,
	 * otherwise the ports must match.
	 */
	for(ctlr = bgbectlrhead; ctlr != nil; ctlr = ctlr->next){
		if(ctlr->active)
			continue;
		if(edev->port == 0 || edev->port == ctlr->port){
			ctlr->active = 1;
			break;
		}
	}
	if(ctlr == nil)
		return -1;

	edev->ctlr = ctlr;
	edev->port = ctlr->port;
	edev->irq = ctlr->pcidev->intl;
	edev->tbdf = ctlr->pcidev->tbdf;
	edev->mbps = 100;
	memmove(edev->ea, ctlr->ea, Eaddrlen);

	/*
	 * Linkage to the generic ethernet driver.
	 */
	edev->attach = bgbeattach;
	edev->transmit = bgbetransmit;
	edev->interrupt = bgbeinterrupt;
	edev->ifstat = bgbeifstat;
	edev->ctl = nil;

	edev->arg = edev;
	edev->promiscuous = bgbepromiscuous;
	edev->multicast = bgbemulticast;

	return 0;
}

void
etherbgbelink(void)
{
	addethercard("bgbe", bgbepnp);
}
